#ifndef ACTION_H
#define ACTION_H

#include "HIDMouseKeyboard.h"
#include "TimerHandler.h"
#include "List.h"



//************************** Interface for Input *******************

class Action_Input {
//    virtual void setValue(double p_Value) = 0;
    virtual double readValue() = 0;
};

//************************** Interface for Output *******************
class Action_Output {
  public:
  virtual void writeValue(double p_Value) = 0;
};


//************************** Action Base Class*******************

/** Base class for Actions.
* Actions can be run manually with a double value as parameter, or can be added to an Input instance, which may run all added actions, when the input's value is changed.
* Action implements the TimerListener and can be used pseudo parallel.
* .
* Derived classes has to implement the following functions:
* bool run(double p_Value) this can be called manually or by Input's runActions().
* void eventTimer(unsigned long p_ActualMillis) this is called by the TimerHandler whenever the TimerHandler::refreshTime() is called and the time is changed.
* This can be used to implement pseudo parallel code by checking the time difference between a startTime (the action started run()) and the actual time.
* To get the time difference use TimerHandler::getTimeDiff_Millis(startTime), this will return the correct time difference even when the timer counter had an overflow.
*/
class Action : public TimerListener {

  public:
    static List<Action>*  s_List_Actions;
    static bool addAction(Action* p_Action);
    static bool removeAction(Action* p_Action);
    static int  getActionCount();
    static Action* getAction(int p_Index);
    //closes the instance and frees memory,
    //returns NULL
    static Action* deleteAction(Action* p_Action);

  public:
    Action();
    virtual ~Action();
    virtual bool run(double p_Value_In) = 0;
    virtual void eventTimer(uint64_t) = 0;
};


//************************** Action_Test*******************

/** Action_Test is for testing purpose. A String and the value is printed to Serial
*/
class Action_Test : public Action {
  public:
    Action_Test(String p_String);
    bool run(double p_Value_In);
    virtual void eventTimer(uint64_t p_ActualMicros);
  protected:
    String m_String;
};


//************************** Action_Blink*******************

/**
* Action for blinking an output.
* The action is done pseudo parallel by implementing TimerHandler's TimerListener.
*/
class Action_Blink : public Action {
  public:
    //Start action when input is closed, stop action when input is opened
    static const int MODE_START_INLOW_STOP_INHIGH = 0;
    //Start action when input is closed and stop action when input is closed again
    static const int MODE_STARTSTOP_INLOW = 1;
    //Start action when input is opened and stop action when input is opened again
    static const int MODE_STARTSTOP_INHIGH = 2;
    /**
     * Run action sequence once when input is closed
     * To run the sequence the Repeats has to > 0
    */
    static const int MODE_RUNONCE_INLOW = 3;
    /**
     * Run action sequence once when input is opened
     * To run the sequence the Repeats has to > 0
    */
    static const int MODE_RUNONCE_INHIGH = 4;
    /**
     * Run action sequence once when input is closed and run the sequence again when input is opened
     * To run the sequence the Repeats has to > 0
    */
    static const int MODE_RUNONCE_INHIGHLOW = 5;
    
  protected:
    Action_Output*  m_Output = NULL;
    int m_OutputValue = LOW;
    unsigned long   m_TimeHigh = 10;
    unsigned long   m_TimeLow = 10;
    int             m_Mode = MODE_STARTSTOP_INLOW;
    int             m_Repeats = 1;

    bool          m_ShouldRun = false;
    bool          m_Started = false;
    int           m_ActualRepeats = 0;
    uint64_t m_LastSwapMillis = 0; //last time swapping output
    unsigned long m_ActualDelay = 0;

    
  public:
  /**
  * Constructor.
  * @param p_Output Pointer to Output instance
  * @param p_TimeHigh duration ms the Output is HIGH
  * @param p_TimeLow duration ms the Output is LOW
  * @param p_Repeats amount of repeats.
  *    p_Repeats > 0 amount of blinking sequences.
  *    p_Repeats == 0 no sequence is played.
  *    p_Repeats < 0 The sequence is played infinitely.
  * @param p_Mode the mode of Action_Blink.
  * @see MODE_START_INLOW_STOP_INHIGH, MODE_STARTSTOP_INLOW, MODE_STARTSTOP_INHIGH, MODE_RUNONCE_INLOW, MODE_RUNONCE_INHIGH, MODE_RUNONCE_INHIGHLOW
  */
    Action_Blink(Action_Output* p_Output, unsigned long p_TimeHigh, unsigned long p_TimeLow ,int p_Repeats, int p_Mode);
    ~Action_Blink();
    //Runs the action, depending on m_Mode p_Value is used to decide which part of a blinking sequence shouöd start
    bool run(double p_Value_In);
    /** Is used to check if durations to swap output are resched 
    * CAUTION: This function may be called inderectly by an interrupt service routine -> don't use delay functions or waste time.
    */
    void eventTimer(uint64_t p_ActualMicros);
  protected:
    void start(); //internaL used by run()
    void stop();  //internaL used by run()
    void runOnce();  //internaL used by run()
};

//************************** Action_HID*******************

/** Action_HID can be used to simulate a Mouse / Keyboard
*/
class Action_HID: public Action {
  protected:
    uint8_t m_Key = 0;
    int m_MouseButton = -1;
    HIDMouseKeyboard* m_HIDMouseKeyboard = NULL;
  public:
    Action_HID();
    Action_HID(uint8_t p_Key, uint8_t p_MouseButton);
    void setKeyboardMouseEvent(uint8_t p_Key, uint8_t p_MouseButton);
    bool run(double p_Value_In);
    virtual void eventTimer(uint64_t p_ActualMicros);

};

//************************** Action_Serial *******************
/**
 * Interface Action_Serial Input
 * A class which implements this interface can be set to a Action_Serial instance,
 * which thean sends the interface message instead of a local message
 */
class Action_Serial_InputInterface {
  public:
    /**
     * Get the message buffer
     */
    virtual char* getMessageSerial_Buffer() = 0;
    /**
     * Get the message length
     */
    virtual size_t getMessageSerial_Length() = 0;
};

class Action_Serial : public Action {
  public:
    static const int  MESSAGEBUFFER_LENGTH = 512;


  protected:
    char m_MessageBuffer[MESSAGEBUFFER_LENGTH];
    size_t m_MessageLength = 0;
    Action_Serial_InputInterface* m_Action_Serial_Input = NULL;
  public:
    Action_Serial();
    ~Action_Serial();
    void setAction_Serial_Interface(Action_Serial_InputInterface* p_Action_Serial_Input);
    void setMessage(char* p_MessageBuffer, size_t p_MessageLength);
    void setMessage(String p_MessageString);
    bool run(double p_Value_In);
    virtual void eventTimer(uint64_t p_ActualMicros);


};

#endif