#ifndef INOUTPUTHANDLER_H
#define INOUTPUTHANDLER_H


#include <Arduino.h>

#include "Action.h"
#include "List.h"


//*****************************************************
//************************** Input *******************
//*****************************************************

/**
* Input is an abstract base class which can be used to derive special input classes (digital, analog, RFID, message receiver,..).
* Input implements all functionality for debouncing an Input. For this purpose Inputs implements the TimerListener interface.
* Action instances can be added to an Input instance, which can be started by runActions(). For this purpose Input implements the interface Action_Input (readValue()).
* The actual value of the input is used for calling Action::run(double).
* 
* Classes that inherit Input must implement the following virtual functions:
* + String getClassName(). This methode should return the name of the class.
*   C has no reflections therefore the class name can be used to decide which type of class the Input instance is .
* + double readValue() to read the actual value of the input. This method is called automatically when checkInput() is called.
*   readValue should return a double value.
*   Derived Input classes which handles other values than double may implement additional methods for there specific value and
*   return a virtual value (e.g. HIGH,LOw) when the value has changed.
*   e.g. A class which receives messages, may return LOW when a message is available(readen) and HIGH when no message is available.
*   
* When a hardware input is interruptable the derived class should attache isr_EventInterrupt to the interrupt pin and call setInterruptable(true).
* CAUTION: Whenever the changed value is handled, the method processed() has to called, this resets the value changed flag and the input is ready for the next value change.
* The methode processed() may be overwritten by a derived class to reset the special input's members.
* @see TimerHandler, TimerListener
* @see Action, Action::run(double)
*/
class Input : public Action_Input, public TimerListener {
  //--- static members ---
  public:
    /** To avoid bouncing the hardware value changes are not handled inside this time
    */
    static const unsigned long DEFAULT_DEBOUNCE_TIMESPAN = 50; //ms
    static bool                 USE_INTERRUPTS;
  protected:
    /**static List of all inputs
    */
    static List<Input>* s_List_Inputs; //static list of all Input instances
    /** Is called automatically by the Input's constructor.
    */
    static bool addInput(Input *p_Input); //is called automatically by the Input's constructor
    /** Is called automatically by the Input's destructor
    */
    static bool removeInput(Input *p_Input); //is called automatically by the destructor

    static bool s_AllInterruptable;  //true = all inputs are interruptable and have isr_EventInterrupt() attached
    static bool checkAllInterruptable(); //check if all Inputs are interruptable, is called automatically when setInterruptable is called
    
  public:
    /**Returns the amount of listet Input instances
    */
    static int  getInputCount();
    /** Returns the INput instance at position p_Index.
    * If p_Index is out of bounds NULL is returned.
    */
    static Input* getInput(int p_Index);
    /** Deletes (remove from list and free memory) of p_Input
    */
    static Input* deleteInput(Input* p_Input);

    /** sets if interrupts are usable for all inputs.
    * @param p_UseInterrupts true = interrupts are used when the hardware input has an interruptable interrupt number. The isr_EventInterrupt is attached.
    * false = interrupts are not used, no isr routine will be attached.
    * Caution: Digital Inputs may bounce very often, when time critical hardwares like rf (Lora, wifi, ..) or other are used, interrupts may interfere the time critical hardware routines.
    */
    static void setUseInterrupts(bool p_UseInterrupts);
    /** Returns if all Inputs are interruptable.
    */
    static bool getAllInterruptable();

    /** The Interrupt Service Routine of the listet interruptable Inputs.
    * Whenever one of the liset inputs is interrupted, this routine will call checkInput() of all listet Inputs.
    */
    static void isr_EventInterrupt(); //interrupt service routine must be static without parameters and is used by all Inputs where the isr is attached


  protected:
    List<Action>* m_List_Actions = new List<Action>(); //list containining all added action which will can be run
    bool          m_ValueChanged = false; //true = the value has changed in the meantime
                                          //false = the value has not changed or the change has been handled (e.g. when runAction() was called)

    double        m_Value = HIGH; //actual set value of the INput, is set by checkValueChanged()
                                  //intial HIGH because Input opened = HIGH, Input closed = LOW
    double        m_ValueReadOld = m_Value;

    bool          m_DebounceOn = false; //true = the Input is debounced
    unsigned long m_DebounceTimeSpan = DEFAULT_DEBOUNCE_TIMESPAN;

    uint64_t      m_LastTimerEvent = 0; //the last time (ms) used in timerEvent() to decide if value changes are bouncing 

    bool          m_IsInterruptable = false; //true = the Input is interruptable and isr_EventInterrupt is attached 
    int           m_Interrupt_Counter = 0; //flag indicating whether an interrupt is occured or checkInput() is manually called 

  public:
    /** default constructor must be called by the constructor of the derived Input classes MyInputClass():Input() 
    */
    Input(); //default constructor must be called by the constructor of the derived Input classes MyInputClass():Input() 
    /** destructor is automatically called when a derived Input class is deleteed.
    */
    virtual ~Input(); //destructor is automatically called when a derived Input class is deleteed.
    /**
    * Returns the name of the derived class.
    * C++ has no reflection, The name can be used to find out which class type it is.
    */
    virtual String getClassName() = 0;

    //--- interrupt and debounce handling ---
  public:
    /**
    * Checks if the input has changed.
    * Called either by interrupt isr or manually.
    * Call this function instead of readValue() to use the debouncing option.
    */
    void checkInput();
    /**
    * Called by the TimerHandler, whenever the TimerHandler's refreshTime() is called and the time has changed.  
    * CAUTION: This function may be called inderectly by an interrupt service routine -> don't use delay functions or waste time.
    * @see TimerHandler, TimerListener
    */
    void eventTimer(uint64_t p_ActualMicros);
    /** Switches the debouncing option on or off.
    */
    void setDebounceOn(bool p_DebounceOn);
    bool getDebounceOn();
    /** sets the time span while changes are declared as bouncing.
    */
    void setDebounceTimeSpan(unsigned long p_DebounceTimeSpan);
    unsigned long getDebounceTimeSpan();

    /** Returns if the Input is interruptable.
    * Attaching the isr_EventInterrupt to the specific interrupt pin has to be done in the derived class. if so setInterruptable(true) has to be called.
    */
    bool isInterruptable();
  protected:
    /** Should be called by the derived class with p_IsInterruptable = true, when isr_EventInterrupt is attached to the inputs interrupt pin 
    * Caution the interrupt pin is not the hardware pin number - use digitalPinToInterrupt(HardwarePinNumber).
    * @see isr_EventInterrupt
    */
    void setInterruptable(bool p_IsInterruptable);
  
  
    //--- Action handling ---0
  public:
    /** Advice a pointer of an Action instance to this Input instance.
    */
    void addAction(Action* p_Action);
    /**  Returns the pointer to the advised Action instance
    */
    Action* getAction(int p_Index);
    /** If Actions is advised, the Action's run() is called with the actual Input's value as parameter.
    * runActions() may be overwritten by derived classes.
    */
    virtual bool runActions();

  //--- value handling ---
  protected:
    /** Is caled automatically when checkInput() is called. 
    */
    bool checkValueChanged();

    /** Has to be implemented by the derived class. This function is called automatically when checkValueChanged() is called.
    * Should return the actual value without setting the value for the class instance, this has to be done by checkValueChanged.
    * Derived classes which handles other values than double, should return LOW, when the input value has changed and HIGH when the input value is not changed.
    */
    virtual double readValue() = 0;

  public:
    /** Should be called by inputs where the input events can occur very quickly and the value may not change between two checkValueChanged() calls.
    * Example Input_Serial can receive several messages very quickly, so that the return value of readValue does not change from one checkValueChanged call to the next.
    * In the other cases this is called automatically by checkValueChenged().
    */
    virtual void setValueChanged(bool p_Changed);
    /** Retuns if the actual value of the input is declared as changed or not.
    */
    bool getValueChanged();

    /**
    * Should be called when input with changed value was handled.
    * This methode can be overwritten by derived classes, but the base's methode has also be called Input::processed().
    */
    virtual void processed();

    /** Is called automatically by checkValueChanged().
    * Normally this is not manaually called.
    */
    void setValue(double p_Value); 
    /** Returns the actual Input's value.
    * Caution: This may be not the actual hardware's value, but the value when checkValueChanged() returns true.
    * In case of digital Inputs the value is HIGH when the input is open and LOW when the input is closed.
    */
    double getValue();
};





//************************** Digital Input *******************

/**
* Input_Digital can be used for digital hardware inputs.
* This class will check if the Pin (m_PinNumber) is interruptable or not. If so the Input's isr_EventInterrupt is attached automatically to the specific intterrupt pin.
* Debouncing is 'on' by default.
*/
class Input_Digital: public Input {
  public:
    static const String CLASSNAME;

  protected:
    int m_PinNumber = -1;
  public:
    /** Constructor.
    * @param p_PinNumber: the hardware's pin number 
    */
    Input_Digital(int p_PinNumber); //adds the instance to the Input's list
    ~Input_Digital(); //removes the instance from the Inputs' list

    /** Returns "Input_Digital".
    */
    String getClassName();
    /** Retrns the hardware pi number
    */
    int getPinNumber();

    /**
    * Reads the digital value of the hardware pin.
    * @see Input::readValue
    */
    double readValue();
};

//************************** Digital Input *******************

/**
* Input_Digital can be used for digital hardware inputs.
* This class will check if the Pin (m_PinNumber) is interruptable or not. If so the Input's isr_EventInterrupt is attached automatically to the specific intterrupt pin.
* Debouncing is 'on' by default.
*/
class Input_Serial: public Input, public Action_Serial_InputInterface {
  public:
    static const String CLASSNAME;
    static const int  MESSAGEBUFFER_LENGTH = 512;

  protected:
    String  m_LastMessage = "";
    char    m_MessageBuffer[MESSAGEBUFFER_LENGTH];
    size_t     m_MessageLength = 0;

  public:
    /** Constructor.
    * @param p_PinNumber: the hardware's pin number 
    */
    Input_Serial(); //adds the instance to the Input's list
    ~Input_Serial(); //removes the instance from the Inputs' list

    /** Returns "Input_Digital".
    */
    String getClassName();

    /**
    * Reads the digital value of the hardware pin.
    * @see Input::readValue
    */
    double readValue();

    char*     getMessageSerial_Buffer();
    size_t    getMessageSerial_Length();
};



//*****************************************************
//************************** Output *******************
//*****************************************************

/**
* The abstract base class for Outputs.
* This class can be used e.g. for Actions. Therefore it implements the interface Action_Output: writeValue(double).
* A derived class has to implement following functions:
* + void writeValue(double p_Value)
* + String getClassName()
*/
class Output : public Action_Output {
  public:
  protected:
    /**
    * List of all Output instances.
    */
    static List<Output>* s_ListOutputs;
    static bool addOutput(Output *p_Output); //called by the constructor
    static bool removeOutput(Output *p_Output); //called by the destructor
  public:
    static int  getOutputCount();
    static Output* getOutput(int p_Index);
    /**closes p_Output frees memory and removes it from the Output's list.
    returns NULL
    */
    static Output* deleteOutput(Output* p_Output);


  public:   
    Output();
    /**
    * closes the instance and removes it from the s_Outputs list
    * to free the emory deleteOutput should be used instead
    */
    virtual ~Output();
    /**
    * Returns the name of the derived class
    * because c has no reflection, the name can be used to determine the class type
    */
    virtual String getClassName() = 0;
    /**
    * sets the output value
    * This function has to be declared in derived classes of Output
    */
    virtual void writeValue(double p_Value) = 0;
    /**
    * Returns the actual written value
    */
    double getValue();

  protected:
    double   m_Value = 0;

};


//************************** Digital Output *******************
class Output_Digital: public Output {
  protected:
    int m_PinNumber = -1;
  public:
    /**
    * Constructor of Output_Digital
    * @param p_HardwarePinNumber the number of an hardware output pin
    */
    Output_Digital(int p_HardwarePinNumber);
    /**
    * closes the instance and removes it from the s_Outputs list
    * to free the emory deleteOutput should be used instead
    * @see #Output::deleteOutput
    */
    ~Output_Digital();
    /**
    * Returns the name of the class
    * because c has no reflection, the name can be used to determine the class type
    */
    String getClassName();
    /**
    * Returns the hardware pin number
    */
    int getPinNumber();
    /**
    * Set the value for digital output
    * @param p_Value < 0.5 LOW will be set for the hardware pin
    * if p_Value >= 0.5 HIGH will be set for the hardware pin
    * @see #getPinNumber
    */
    void writeValue(double p_Value);
};

#endif