#include "wiring_constants.h"
#include "InOutputHandler.h"


//**************************************************************
//************************** Input Base Class*******************
//**************************************************************


//--- static ---
//initialize static members

List<Input>* Input::s_List_Inputs = new List<Input>(20); //create list with initial capacity of 20 
bool Input::USE_INTERRUPTS = false;
bool Input::s_AllInterruptable = false;


bool Input::addInput(Input *p_Input) { 
  return s_List_Inputs->addElement(p_Input);
}
bool Input::removeInput(Input* p_Input) {
  if (s_List_Inputs->removeElement(p_Input) == false) {
    return false;
  }
  //check all interruptable when p_Input is not interruptable
  if (p_Input->isInterruptable() == false) {
    checkAllInterruptable();
  }
  return true;
}
int Input::getInputCount() {
  return s_List_Inputs->getSize();
}
Input* Input::getInput(int p_Index) {
  return s_List_Inputs->getElement(p_Index);
}

Input* Input::deleteInput(Input *p_Input) {
  if (p_Input == NULL) return NULL;
  delete p_Input;
  return NULL;
}

void Input::setUseInterrupts(bool p_UseInterrupts) {
  USE_INTERRUPTS = p_UseInterrupts;
}

bool Input::checkAllInterruptable() {
  Input* ptrInput;
  bool boolAllInterruptable = true;
  for (int intInput = 0; intInput < s_List_Inputs->getSize(); intInput++) {
    ptrInput = getInput(intInput);
    if (ptrInput == NULL) continue;
    if (ptrInput->isInterruptable() == false) {
      boolAllInterruptable = false;
      break;
    }
  }
  s_AllInterruptable = boolAllInterruptable;
  return boolAllInterruptable;
}
bool Input::getAllInterruptable() {
  return s_AllInterruptable;
}


//--- non static ---
Input::Input()  {
  addInput(this);
  TimerHandler::addTimerListener(this);
}
Input::~Input() {
  TimerHandler::removeTimerListener(this);
  m_ValueChanged = false;
  removeInput(this);
}

void Input::setInterruptable(bool p_IsInterruptable) {
  m_IsInterruptable = p_IsInterruptable;
  //check allInterruptable
  checkAllInterruptable();
}
bool Input::isInterruptable() {
  return m_IsInterruptable;
}

//interrupt service routine
void Input::isr_EventInterrupt() {
  Input* ptrInput;
  for (int intInput = 0; intInput < getInputCount(); intInput++) {
    ptrInput = getInput(intInput);
    if (ptrInput == NULL) continue;
    if (ptrInput->isInterruptable()) {
      ptrInput->checkInput();
    }
  }
}

void Input::checkInput() {
  // 1.) Whenever an input is changed the methode isr_IterruptEvent will call checkInput() for every input
  //     -> check if this input is changed
  // 2.) checkInput maybe called manually to check if the input is changed
  if (m_Interrupt_Counter <= 1) {
    m_Interrupt_Counter++; //The interrupt could have been triggered by this input
  }

}

void Input::eventTimer(uint64_t p_ActualMicros) {
  uint64_t Uint64TimeDiff;
  
  if (m_Interrupt_Counter <= 0) {
    //no interrupt occured
    return;
  }

  Uint64TimeDiff = TimerHandler::getTimeDiff_Millis(m_LastTimerEvent);
  if (m_DebounceOn && Uint64TimeDiff < m_DebounceTimeSpan) {
    if (m_Interrupt_Counter > 1) {
      //decrement m_Interrupt_Counter only to 1 to be shure the next eventTimer will call checkValueChanged
      m_Interrupt_Counter--;
      return;
    }
  } 
  bool boolValueChanged;
  boolValueChanged = checkValueChanged();
  if (boolValueChanged == true) {
    m_ValueChanged = true;
  }
  
  if (m_ValueChanged == true) {
    m_LastTimerEvent = TimerHandler::getTime_ActualMillis();
    // set m_Interrupt_Counter to 1 to reduce the amount of eventTimer handling
    // set m_Interrupt_Counter NOT to 0, because in the meantime an interrupt may occure 
    m_Interrupt_Counter = 1; 
  }
  else {
    //decrement m_Interrupt_Counter
    if (m_Interrupt_Counter > 0) m_Interrupt_Counter --;
  } 
  return;
}

void Input::setDebounceOn(bool p_DebounceOn) {
  m_DebounceOn = p_DebounceOn;
}
bool Input::getDebounceOn() {
  return m_DebounceOn;
}
void Input::setDebounceTimeSpan(unsigned long p_DebounceTimeSpan) {
  m_DebounceTimeSpan = p_DebounceTimeSpan;
}
unsigned long Input::getDebounceTimeSpan() {
  return m_DebounceTimeSpan;
}

void Input::addAction(Action* p_Action) {
  m_List_Actions->addElement(p_Action);
}
Action* Input::getAction(int p_Index) {
  return m_List_Actions->getElement(p_Index);
}
bool Input::runActions() {
  bool boolResult = true;
  Action* ptrAction;
  for (int intAction = 0; intAction < m_List_Actions->getSize(); intAction++) {
    ptrAction = m_List_Actions->getElement(intAction);
    if (ptrAction->run(m_Value) == false) boolResult = false;;
  }
  return boolResult;
}


void Input::setValue(double p_Value) {
  m_Value = p_Value;
}
double Input::getValue() {
  return m_Value;
}

void Input::setValueChanged(bool p_Changed) {
  m_ValueChanged = p_Changed;
}
bool Input::getValueChanged() {
  return m_ValueChanged;
}

void Input::processed() {
  setValueChanged(false);
}

bool Input::checkValueChanged() {
  double doubleValue;
  doubleValue = readValue();
  if(doubleValue == m_Value) return false;
  setValue(doubleValue);
  setValueChanged(true);
  return true;
}




//************************** Input_Digital *******************
const String Input_Digital::CLASSNAME = "Input_Digital";

Input_Digital::Input_Digital(int p_PinNumber) : Input() {
  if (p_PinNumber >= 0) {
    m_PinNumber = p_PinNumber;
    
    //set pinMode to INPUT
    pinMode(m_PinNumber, INPUT_PULLUP);

    setDebounceOn(true); //switch debounce on

    //check interruptable
    setInterruptable(false);
    if (USE_INTERRUPTS == true) {
      int intInterruptNumber;
      intInterruptNumber = digitalPinToInterrupt(m_PinNumber);
      if (intInterruptNumber != NOT_AN_INTERRUPT) {
        //interrupt is possible -> attach interrupt methode
        attachInterrupt (intInterruptNumber, isr_EventInterrupt, CHANGE);  // attach interrupt handler
        setInterruptable(true);
      }
    }
  }
}
Input_Digital::~Input_Digital() {
  if (m_IsInterruptable == true) {
    //check interruptable
    int intInterruptNumber;
    intInterruptNumber = digitalPinToInterrupt(m_PinNumber);
    if (intInterruptNumber != NOT_AN_INTERRUPT) {
        //detach interrupt service routine
        detachInterrupt(intInterruptNumber);  // detach interrupt handler
        setInterruptable(false);
    }
  }
  m_List_Actions->clear();
}

double Input_Digital::readValue() {
  return digitalRead(m_PinNumber);
}


String Input_Digital::getClassName() {
  return CLASSNAME;
}
int Input_Digital::getPinNumber() {
  return m_PinNumber;
}

//************************** Input_Digital *******************
const String Input_Serial::CLASSNAME = "Input_Serial";

Input_Serial::Input_Serial() : Input() {
  m_MessageLength = 0;
  m_LastMessage = "";
}
Input_Serial::~Input_Serial() {

  m_List_Actions->clear();
}
String Input_Serial::getClassName() {
  return CLASSNAME;
}

double Input_Serial::readValue() {
  int intAvallable;
//  if (!Serial) {
//    return HIGH;
//  }
  if (!Serial) {
    return HIGH;
  }
  intAvallable = Serial.available();
  if (intAvallable <= 0) {
    return HIGH;
  }

  
  m_MessageLength = Serial.readBytes(m_MessageBuffer, intAvallable);
  if (m_MessageLength > 0) {    
  
    setValueChanged(true);  //setValueChanged to true because a new message arrived.
                            //and the last time readValue was called, a new message may also have arrived. -> the last returnd vaue was also LOW

    return LOW;
  }

  return HIGH;
}
char* Input_Serial::getMessageSerial_Buffer() {
  return m_MessageBuffer;
}
size_t Input_Serial::getMessageSerial_Length() {
  return m_MessageLength;
}







//***************************************************************
//************************** Output base class*******************
//***************************************************************

//--- static ---

//initialize static members
List<Output>* Output::s_ListOutputs = new List<Output>(25);

bool Output::addOutput(Output *p_Output) {
  return s_ListOutputs->addElement(p_Output);  
}
bool Output::removeOutput(Output *p_Output) {
  return s_ListOutputs->removeElement(p_Output);
}
int Output::getOutputCount() {
  return s_ListOutputs->getSize();
}
Output* Output::getOutput(int p_Index) {
  return s_ListOutputs->getElement(p_Index);
}
Output* Output::deleteOutput(Output *p_Output) {
  if (p_Output == NULL) return NULL;
  delete p_Output;
  return NULL;
}

//--- non static ---
Output::Output() {
  addOutput(this);
}
Output::~Output() {
  removeOutput(this);
}

double Output::getValue() {
  return m_Value;
}




//************************** Output_Digital *******************
Output_Digital::Output_Digital(int p_PinNumber):Output() {
  if (p_PinNumber >= 0) {
    m_PinNumber = p_PinNumber;
    
    //set pinMode to OUTPUT
    pinMode(m_PinNumber, OUTPUT);

  }
}
Output_Digital::~Output_Digital() {
}

String Output_Digital::getClassName() {
  return "Output_Digital";
}
int Output_Digital::getPinNumber() {
  return m_PinNumber;
}

void Output_Digital::writeValue(double p_Value) {
  if (p_Value < 0.5) {
    //value is lower than 0.5 -> write LOW
    m_Value = LOW;
    digitalWrite(m_PinNumber, LOW);
  }
  else {
    //value is greater or equal 0.5 -> write HIGH
    m_Value = HIGH;
    digitalWrite(m_PinNumber, HIGH);
  }
}

