
#ifndef LORAHANDLER_H
#define LORAHANDLER_H


#include <SPI.h>
#include <RH_RF95.h>

#include "InOutputHandler.h"
#include "List.h"
#include "StorageHandler.h"



//-------------------- Lora Definitions------------------

#if defined (__AVR_ATmega32U4__)  // Feather 32u4 w/Radio
  #define RFM95_CS    8
  #define RFM95_INT   7
  #define RFM95_RST   4

#elif defined(ADAFRUIT_FEATHER_M0) || defined(ADAFRUIT_FEATHER_M0_EXPRESS) || defined(ARDUINO_SAMD_FEATHER_M0)  // Feather M0 w/Radio
  #define RFM95_CS    8
  #define RFM95_INT   3
  #define RFM95_RST   4

#elif defined(ARDUINO_ADAFRUIT_FEATHER_RP2040_RFM)  // Feather RP2040 w/Radio
  #define RFM95_CS   16
  #define RFM95_INT  21
  #define RFM95_RST  17

#elif defined (__AVR_ATmega328P__)  // Feather 328P w/wing
  #define RFM95_CS    4  //
  #define RFM95_INT   3  //
  #define RFM95_RST   2  // "A"

#elif defined(ESP8266)  // ESP8266 feather w/wing
  #define RFM95_CS    2  // "E"
  #define RFM95_INT  15  // "B"
  #define RFM95_RST  16  // "D"

#elif defined(ARDUINO_ADAFRUIT_FEATHER_ESP32S2) || defined(ARDUINO_NRF52840_FEATHER) || defined(ARDUINO_NRF52840_FEATHER_SENSE)
  #define RFM95_CS   10  // "B"
  #define RFM95_INT   9  // "A"
  #define RFM95_RST  11  // "C"

#elif defined(ESP32)  // ESP32 feather w/wing
  #define RFM95_CS   33  // "B"
  #define RFM95_INT  27  // "A"
  #define RFM95_RST  13

#elif defined(ARDUINO_NRF52832_FEATHER)  // nRF52832 feather w/wing
  #define RFM95_CS   11  // "B"
  #define RFM95_INT  31  // "C"
  #define RFM95_RST   7  // "A"

#endif

// Change to 434.0 or other frequency, must match RX's freq!
#define RF95_FREQ 433.0

//-------------------- LoraMessage ------------------

/** LoraMessage can be used to send over Lora
* Bytes[] = [Precode][Message]
* Precode: [0] = Precode length (inclusive this byte)
*          [1..4] = Sender ID
*          [5] = Sender Type
* Message: [6+++] = message chars has to finsihed with \0 (null terminator)
*/

class LoraMessage {
  public:
    static const int LENGTH_PRECODELENGTH = 1; //length of the precode length. The precode length keeps the length of the precode (0..255)
  
    static const int LENGTH_SENDERID = 4;      //length of the sender ID
    static const int LENGTH_SENDERTYPE = 1;    //length of the sender type 

    static const int  MESSAGESERIAL_LENGTH = 512; //length of the char buffer for serial massages 


    static String VALUE_SEPERATOR;

  public:
    LoraMessage();
    ~LoraMessage();
  protected:
    /**
    * Received bytes or calculated bytes by set-functions (setSenderID, setMessage,...)
    */
    char  m_Bytes[RH_RF95_MAX_MESSAGE_LEN]; //char array keeps the lora message
    int   m_Length = 0; //keeps the complete length of the lora message

    /**
    * For serial sending purpose.
    * can be used to send a representation of the LoraMessage to serial 
    */
    char  m_MessageSerial_Bytes[MESSAGESERIAL_LENGTH];
    int   m_MessageSerial_Length = 0;  
  public:
    /** Returns the maximal possible length of a LoraMessage, depends on Lora.
    */
    int           getMaxLength();


    /** Precode: Set the ID of the sender: (0 .. 4.294.967.295).
    * Should be set to identify a sender
    */
    void          setSenderID(uint32_t p_SenderID);
    /** Precode: Returns the sender ID.
    */
    uint32_t      getSenderID();
    String        getSenderID_Hex();

    /** Precode: Set the type of the sender: 1 char.
    * maybe used to distinguish between different types of senders
    * This actually not used
    */
    void          setSenderType(char p_SenderType);
    /** Precode: Returns the sender type.
     * This is actually not used
    */
    char          getSenderType();
    /** Returns the lemgth of the precode.
    * Inclusive the one byte, containing the precode length
    */
    unsigned int  getPrecodeLength();


    /** 
     * Payload: Set the message and value string.
     * normally the message is a single character 
    */
    void          setMessage(String p_Message, String p_Value);
    /** Returns the whole message inclusive value seperator and value.
    */
    String        getMessage();
    /** Returns the message without value seperator and value.
    */
    String        getMessageWithoutValue();
    /** Returns the value as string
    */
    String        getValueString();
    /** Returns the value.
    */
    double        getValue();
    
    /** Set the whole length (precode + message string) the message string has to end with null terminator '\0'.
    * Is called by either by read() or ba setting the message string setMessage().
    */
    void          setLength(int p_Length);
    /** Returns the whole length of the LoraMessage (precode + message string) the message string has to end with null terminator '\0'.
    */
    int           getLength();

    /** Pointer to the LoraMessage's char array.
    */
    char*         getBytes();
    
    /** Returns the content of the LoraMessage as readable string
    */
    String        getString();
    /** Prints the content of LoraMessage to Serial
    */
    void          printToSerial();

    void          calculateMessageSerial();
    char*         getMessageSerial_Buffer();
    size_t        getMessageSerial_Length();
};

//-------------------- Lora ------------------

/** Static instance of Lora (all methods are static).
* Lora handles:
* init() Initialization of lora.
* send() sending a LoraMessage instance (TX).
* read() check if a LoraMessage is available and reads in the LoraMessage (RX).
*/
class Lora {
  public:
  // Singleton instance of the radio driver
    static RH_RF95      rf95;
    static bool         initCalled;
    static const int    MAX_SENDERIDS = 20;
    static const char   DEFAULT_STARTCODE = 'A';
    static uint32_t*    s_SenderIDs;
    static char*        s_Sender_StartCode;
    static int          s_SenderIDCount;

    //used for saving and reading sender ids
    static int              s_EEPROM_Array_MaxLength;
    static char*            s_EEPROM_Array;
    static int              s_EEPROM_HeaderSize;
    static StorageHandler*  s_StorageHandler; //the class StorageHandler has to be rewritten for specific Flas/EEPROM chips

    //learning mode
    static bool           s_LearningActive; //true = the learning mode is active, received LoraMessages will be used to add the SenderID
    static bool           s_Learning_ChangedSenderID; //true = the ist of sender ids is changed
    static Action_Blink*  s_Action_Blink_Learning; //runs while learning is active

    static Output*        s_Output_LED_IDS_CLeared;

  public:
    /** Initializes Lora, has to be called by Arduinio setup().
    */
    static bool init(Output* p_Output_LED_Response);

    /**
     * Clears the list of added SendeIDs
     */
    static void clearSenderIDs();
    /**
     * Add a SenderID, all message from this Sender will be accepted further
     */
    static bool addSenderID(uint32_t p_SenderID);
    static uint32_t getSenderID(int p_Index);
    static int getSenderID_Index(uint32_t p_SenderID);
    /**
     * SenderID_Startcode is normally DEFAULT_STARTCODE
     * This function can be used to change the StartCode.
     * All message received by the LoraReceiver will be reinterpreted:
     * receivedCode + (newSTartCode - DEFAULT_STARTCODE)
     * @param p_SenderID the id of the sender
     * @param p_StartCode the new start code (char) 
     */
    static bool setSenderID_StartCode(uint32_t p_SenderID, char p_StartCode);
    /**
     * Returns the StartCode for a senderID
     * @param p_SenderID the id of the sender
     * @see setSenderID_StartCode
     */
    static char getSenderID_StartCode(uint32_t p_SenderID);
    static bool containsSenderID(uint32_t p_SenderID);
    static int  getSenderIDCount();

    /**
     * save and read SenderIDs, the StorageHandler uses actually only the chip Winbond SPIFlash 3V 64M-BIT
     * for other Flash/EEprom chips the StorageHandler has to be rewritten
     */
    static void setStorageHandler(StorageHandler* p_StorageHandler);
    static void saveSenderIDs();
    static void readSenderIDs();
    /**
     * The EEPROM_Array is used for saving the list of SenderIDs
     * fillEEPROM_Array will fill the array with the list of SenderIDs + some saving code
     */
    static char* fillEEPROM_Array();
    /**
     * The EEPROM_Array is used for saving the list of SenderIDs
     * interpreteEEPROM_Array will interprete the loaded array and fills the list of SenderIDs
     */
    static bool interpreteEEPROM_Array();

    //Learning
    /**
     * Sets the learning mode active or inactive.
     * When learning mode is active a received message from a sender will be used to add the sender's ID.
     * When learning mode is inactive a recieved message will be interpreted the normal way.
     */
    static void setLearningActive(bool p_LearningActive);
    static bool getLearningActive();


    /** sends a LoraMessage
    */
    static bool send(LoraMessage* p_LoraMessage);
    /** Check if a LoraMessage is available.
    */
    static bool getAvailable();
    /** Tries to read eads a available LoraMessage.
    * @return -1 = no LoraMessage available
    * @return > 0 = the length of the readen LoraMessage
    */
    static int  read(LoraMessage* p_LoraMessage);
    /** Returns a receiver signal strength indicator.
    * This number will range from about -15 to about -100. 
    * The larger the number (-15 being the highest you'll likely see) the stronger the signal.
    */
    static int  getastRSSI();
};



//-------------------- Input_Lora ------------------

/** 
 * Input_Lora can be used to receive LoraMessages.
* If a LoraMessage is available, the message is readen and the Inputs's value is set to LOW.
* If no message available the Inputs's value is set to HIGH.
* If an action is assigned and a LoraMessage is readen, the action will run twice: 
* run(LOW) LoraMessage readen.
* run(HIGH) LoraMessage handled (Input processed). 
* With setSendResponse(true) a copy of the received LoraMessage is send back to the sender(default false). 
*/
class Input_Lora : public Input, public Action_Serial_InputInterface {
  public:
    Input_Lora();
    ~Input_Lora();
    String        getClassName(); //returns "Input_Lora"
    double        readValue(); //tries to read a LoraMessage
    bool          runActions();

    /** set if an response should be send when a message is received or not. Default is false.
    * p_SendResponse true = a response is sended. A copy of the received message is sended back to the sender.
    * false = no response is sended. 
    */
    void          setSendResponse(bool p_SendResponse);
    /** Returns a pointer to the last readen LoraMessage.
    */
    LoraMessage*  getLoraMessage();
    void          changeMessageStartCode();
    /**
     * Returns a char array from LoraMessage
     */
    char* getMessageSerial_Buffer();
    /**
     * Returns the length of the LoraMessage
     */
    size_t getMessageSerial_Length();

  protected:
    bool          m_SendResponse = false;
    LoraMessage*  m_LoraMessage_Received;
};

//-------------------- Action_Lora_Send ------------------

/** 
 * Action_Lora_send can be used to send a LoraMessage as a reaction of an Input change.
 * The LoraMessage contains:
 * + the the sender ID, the sender type as set by the constructor.
 * + the message string as set by the constructor. Can be changed by setMessage()
 * + the value seperator char (default ':'). Can be changed by setValueSeperator().
 * + the value with decimal places (default 0). Can be canges by setValueDecimalPlaces().
*/
class Action_Lora_Send : public Action {
  public:
    static const char DEFAULT_VALUESEPERATOR = ':'; //":"
    static const int DEFAULT_VALUEDECIMALPLACES = 0; 

  public:
    /**
     *  Constructor
     * @param p_SenderID = the id of the sender value from 0 to 4.294.967.295.
     * @param p_SenderType = a char which can be used to distinguish between different types of senders
     * @param p_Message = the message to send exclusive the value
    */
    Action_Lora_Send(uint32_t p_SenderID, char p_SenderType, String p_Message);
    ~Action_Lora_Send();
    /** Set the message sended by run(value) without the value
    */
    void setMessage(String p_MessageWithoutValue);
    /** Set the decimal places of the value diplayed in the LoraMessage.
    * if p_DecimalPlaces < 0, no value will be transmitted.
    * 0 means the value is senden without without decimal places.
    */
    void setValueDecimalPlaces(int p_DecimalPlaces);
    /** run() will send a LoraMessage with the given value.
    * The decimal places is used to decide how many decimal places of the value is sended,
    * or if the value is attached.
    */
    bool run(double p_ValueIn);

    void eventTimer(uint64_t p_ActualMicros);
    
  protected:
    LoraMessage*  m_LoraMessage; //message for value LOW in run()
    String        m_MessageWithoutValue  = "";
    char          m_ValueSeperator = DEFAULT_VALUESEPERATOR; //seperator
    int           m_ValueDecimalPlaces = DEFAULT_VALUEDECIMALPLACES; //decimal places of the value 
};

//-------------------- Action_Lora_Received ------------------
/** Action_Lora_Received can be used to as a response when a LoraMessage is received
* When Serial is connected and a Lora_Input is set a representation of the LoraMessage is written by Serial.
* If Serial is not connected the LoraMessage's message is used to call a HID Keyboard event.
*/
class Action_Lora_Received : public Action_Serial {
  public:
    static const char DEFAULT_VALUESEPERATOR = ':'; //":"
    static const int DEFAULT_VALUEDECIMALPLACES = 0; 

  public:
    /** Constructor
    */
    Action_Lora_Received();
    ~Action_Lora_Received();
    /** Stes the Input_Lora which received LOraMessage is used to send serial or HID event.
    */
    void setInput_Lora(Input_Lora* p_Input_Lora);
    bool run(double p_ValueIn);

//    void eventTimer(unsigned long p_ActualMicros);
  protected:
    Input_Lora* m_Input_Lora = NULL; //the Input_Lora which received LoraMessage is used to send serial or HID event
    Action_HID* m_Action_HID; //the HID action which is sused to send a HID event, when Serial is not sconnected
    
};

/** Action_LoraLearning will start the learning process of a Lora receiver.
* When Lora's learning is active the received LoraMessages are used to add allowed sender id's.
* An Output (m_Output_ResponseClear) can be definded to indicate when the sender ids are cleared.
*/
class Action_LoraLearning : public Action {
  public:
    static const int TIME_CLEARSERVERIDS_MS = 3000;
    static const int TIME_OUTPUTON_AFTERCLEAR_MS = 1000;
  public:
    Action_LoraLearning(Output* p_Output_LED);
    ~Action_LoraLearning();
    bool run(double p_Value_In);
    void eventTimer(uint64_t p_ActualMicros);
  protected:
    bool  m_ActionStarted = false;
    bool  m_ShouldRun = false;
    bool  m_LearningActive = false;
    uint64_t m_StartTime;
    bool  m_Clear_IsDone = false;
    Output* m_Output_ResponseClear = NULL;
};

/**
 * Action_Serial_ChangeStartCode is added to an Input_Serial.
 * When a serial message like "LoraSender(SenderID)_X" is received,
 * the Lora Receiver will interprete a received LoraMessage's messageWithoutValue:
 * messageWithoutValue + (charStartCode - Lora::DEFAULT_STARTCODE)
 */
class Action_Serial_ChangeStartCode : public Action {
  public:
    static String COMMAND_START;
  public:
    Action_Serial_ChangeStartCode();
    ~Action_Serial_ChangeStartCode();

    Action_Serial_InputInterface* m_Action_Serial_InputInterface = NULL;
    /**
     * Sets the Action_Serial_InputInterface, which is an Input_Serial
     */
    void setAction_Serial_InputInterface(Action_Serial_InputInterface* p_Action_Serial_InputInterface);
    bool run(double p_Value_In);
    void eventTimer(uint64_t p_ActualMicros);

};

#endif